var MIN_ATTACKERS = 12;
var MIN_NEXUS = 4;
var MAX_SENSORS; // initialized at game start
var MAX_DEFENDERS = 8;
var MAX_GLOBAL_DEFENDERS = 25;

// unit limit constant
function atLimits() {
	return enumDroid(me).length >= 130;
}

// random integer between 0 and max-1 (for convenience)
function random(max) {
	return (max <= 0) ? 0 : Math.floor(Math.random() * max)
}

// Returns true if something is defined
function isDefined(data) {
	return typeof(data) !== "undefined";
}

function log(message) {
	dump(gameTime + " : " + message);
}

function logObj(obj, message) {
	dump(gameTime + " [" + obj.name + " id=" + obj.id + "] > " + message);
}

// Make sure a unit does not try to go off map
// {x, y} are coordinates of some object.
// num1 and num2 move a coordinate a small amount.
// {xOffset, yOffset} influence that direction more.
function mapLimits(x, y, num1, num2, xOffset, yOffset) {
	var coordinates = [];
	var xPos = x + xOffset + random(num1) - num2;
	var yPos = y + yOffset + random(num1) - num2;

	if(xPos < 2)
		xPos = 2;
	if(yPos < 2)
		yPos = 2;
	if(xPos >= mapWidth - 2)
		xPos = mapWidth - 3;
	if(yPos >= mapHeight - 2)
		yPos = mapHeight - 3;
	
	coordinates.push(xPos);
	coordinates.push(yPos);
	return coordinates;
}

function rangeStep(obj, visibility) {
	const step = 10000;
	var target;
	
	for(var i = 0; i < 30000; i += step) {	
		var temp = enumRange(obj.x, obj.y, i, ENEMIES, visibility);
		if(temp.length > 0) {
			return temp[0];
		}
	}
	
	return target;
}

const derrick = "A0ResourceExtractor";
const factory = "A0BaBaFactory";
const vtolfac = "A0BaBaVtolFactory";
const gen = "A0BaBaPowerGenerator";
const oilres = "OilResource";
const repair = "ScavRepairCentre";
const vtolpad = "A0BaBaVtolPad"

const defenses = [
	"A0BaBaBunker",
	"A0BaBaBunker",
	"A0BaBaBunker",
	"A0CannonTower",
	"A0CannonTower",
	"A0CannonTower",
	"A0BaBaFlameTower",
	"A0BaBaFlameTower",
	"A0BaBaRocketPit",
	"A0BaBaRocketPit",
	"A0BaBaRocketPitAT",
	"A0BaBaMortarPit",
	"bbaatow",
];

//["B1BaBaPerson01","BaBaMG"],
const templates = [
	["B4body-sml-trike01","bTrikeMG"],
	["B4body-sml-trike01","bTrikeMG"],
	["B4body-sml-trike01","bTrikeMG"],
	["B4body-sml-trike01","bTrikeMG"],
	["B4body-sml-trike01","bTrikeMG"],
	["B4body-sml-trike01","bTrikeMG"],
	["B3body-sml-buggy01","BuggyMG"],
	["B3body-sml-buggy01","BuggyMG"],
	["B3body-sml-buggy01","BuggyMG"],
	["B3body-sml-buggy01","BuggyMG"],
	["B2JeepBody","BJeepMG"],
	["B2JeepBody","BJeepMG"],
	["B2JeepBody","BJeepMG"],
	["B2JeepBody","BJeepMG"],
	["B3bodyRKbuggy01","BabaRocket"],
	["B3bodyRKbuggy01","BabaRocket"],
	["B3bodyRKbuggy01","BabaRocket"],
	["B2RKJeepBody","BabaRocket"],
	["B2RKJeepBody","BabaRocket"],
	["B2RKJeepBody","BabaRocket"],
	["BusBody","BusCannon"],
	["BusBody","BusCannon"],
	["BusBody","BabaPitRocketAT"],
	["B2tractor","BabaFlame"],
	["B2tractor","BabaFlame"],
	["B2tractor","BabaFlame"],
	["FireBody","BabaFlame"],
	["FireBody","BabaFlame"],
	["FireBody","BusCannon"],
	["FireBody","BabaPitRocket"],
	["FireBody","BabaPitRocketAT"],
	["ScavCamperBody","BabaPitRocket"],
	["ScavCamperBody","BusCannon"],
	["ScavTruckBody","BabaFlame","BabaRocket","BabaPitRocketAT"],
	["ScavTruckBody","BusCannon","BabaPitRocket","BabaRocket"],
	["ScavIcevanBody","BabaFlame"],
	["ScavIcevanBody","Mortar1Mk1"],
	["ScavNEXUStrack","ScavNEXUSlink"],
	["ScavNEXUStrack","ScavNEXUSlink"],
	["ScavNEXUStrack","ScavNEXUSlink"],
];

const vtolTemplates = [
	["ScavengerChopper","MG1-VTOL","MG1-VTOL","MG1-VTOL"],
	["ScavengerChopper","MG1-VTOL","MG1-VTOL","MG1-VTOL"],
	["HeavyChopper","Rocket-VTOL-Pod"],
];

// scav groups
var globalDefendGroup; // tanks that defend all bases
var needToPickGroup; // a group

var baseInfo = [];

function constructBaseInfo(x, y) {
	this.x = x;
	this.y = y;
	this.defendGroup = newGroup(); // tanks to defend the base
	this.nexusGroup = newGroup(); // tanks to capture the enemy
	this.builderGroup = newGroup(); // trucks to build base structures and defenses
	this.attackGroup = newGroup(); // tanks to attack nearby things
	this.factoryNumber = baseInfo.length;
}

function findNearest(list, x, y, flag) {
	var minDist = Infinity, minIdx;
	for (var i = 0; i < list.length; ++i) {
		var d = distBetweenTwoPoints(list[i].x, list[i].y, x, y);
		if (d < minDist) {
			minDist = d;
			minIdx = i;
		}
	}
	return (flag === true) ? list[minIdx] : minIdx;
}

function reviseGroups() {
	var list = enumGroup(needToPickGroup);
	for (var i = 0; i < list.length; ++i) {
		var droid = list[i];
		addDroidToSomeGroup(droid);
		var coords = mapLimits(droid.x, droid.y, 15, 7, 0, 0);
		orderDroidLoc(droid, DORDER_SCOUT, coords[0], coords[1]); 
	}
}

function addDroidToSomeGroup(droid) {
	var base = findNearest(baseInfo, droid.x, droid.y, true);
	switch(droid.droidType) {
		
		case DROID_CONSTRUCT: {
			groupAddDroid(base.builderGroup, droid);
			break;
		}	
		case DROID_WEAPON: {
			if (droid.name.indexOf("Nexus") > -1) {
				
				if(groupSize(base.nexusGroup) < (2 * MIN_NEXUS))
					groupAddDroid(base.nexusGroup, droid); 
				else {
					var rBase = random(baseInfo.length);
					groupAddDroid(baseInfo[rBase].nexusGroup, droid);
				}
				break;
				
			}
			switch(random(5)) {
				case 0: 
					if (groupSize(globalDefendGroup) < MAX_GLOBAL_DEFENDERS) {
						groupAddDroid(globalDefendGroup, droid); 
						break;
					}
					
				case 1: {
					if (groupSize(base.defendGroup) < MAX_DEFENDERS)
						groupAddDroid(base.defendGroup, droid); 
					else {
						var rBase = random(baseInfo.length);
						groupAddDroid(baseInfo[rBase].defendGroup, droid);
					}
					break;
				}
				
				default: {
					if(groupSize(base.attackGroup) < (2 * MIN_ATTACKERS))
						groupAddDroid(base.attackGroup, droid);
					else {
						var rBase = random(baseInfo.length);
						groupAddDroid(baseInfo[rBase].attackGroup, droid);
					}
					break;
				}
			}
		}
		case DROID_SENSOR: { 
			groupAddDroid(base.attackGroup, droid); 
			break;
		}
		case DROID_PERSON: {
			groupAddDroid(base.defendGroup, droid);
			break;
		}
	}
}

function groupOfTank(droid) {
	for (var i = 0; i < baseInfo.length; ++i) {
		if (droid.group == baseInfo[i].attackGroup)
			return baseInfo[i].attackGroup;
		
		if (droid.group == baseInfo[i].nexusGroup)
			return baseInfo[i].nexusGroup;
	}
}

function buildStructure(droid,stat) {
	if (droid.order == DORDER_BUILD)
		return false;
	
	if (!isStructureAvailable(stat,me))
		return false;
	
	var loc = pickStructLocation(droid, stat, droid.x, droid.y, 0);
	if (!isDefined(loc))
		return false;
	
	orderDroidBuild(droid, DORDER_BUILD, stat, loc.x, loc.y); 
	return true;
}

function buildTower(droid) {
	return buildStructure(droid,defenses[random(defenses.length)]);
}

function establishBase(droid) {
	var base = findNearest(baseInfo, droid.x, droid.y, true);
	var dist = distBetweenTwoPoints(base.x, base.y, droid.x, droid.y);
	
	//dist makes sure that factories are not built too close to eachother
	if ((dist > 8) && buildStructure(droid, factory)) {
		var n = baseInfo.length;
		baseInfo[n] = new constructBaseInfo(droid.x, droid.y);
		groupAddDroid(baseInfo[n].builderGroup, droid);
		logObj(droid, "establishBase: built a new base.");
		return true;
	}
	if(dist > 8)
		logObj(droid, "establishBase: Failed to construct a new base.");
	return false;
}

function buildThingsWithDroid(droid) {
	switch(random(7)) {
	case 0: {
		if (!atLimits() && ((5 * countStruct(factory)) < countStruct(derrick)) || (playerPower(me) > 500))
			establishBase(droid);
		break;
	}
	case 1: {
		if ((countStruct(derrick) - (countStruct(gen) * 4)) > 0)
			buildStructure(droid, gen);
		break;
	}
	case 2: {
		if ((4*countStruct(vtolfac)) < countStruct(factory) && (gameTime > 150000))
			buildStructure(droid, vtolfac);
		break;
	}
	case 3: {
		var result = findNearest(enumFeature(-1, oilres), droid.x, droid.y, true);
		if (isDefined(result))
			orderDroidBuild(droid, DORDER_BUILD, derrick, result.x, result.y);
		break;
	}
	case 4: {
		if ((countStruct(repair) < 5) && (gameTime > 200000)) {
			var reps = countStruct(repair);
			buildStructure(droid, repair);
		}
		break;
	}
	case 5: {
		if (sortHelicopters().length > 2 * countStruct(vtolpad) )
			buildStructure(droid, vtolpad);
		break;
	}
	default: {
		if (playerPower(me)>300) 
			buildTower(droid);
		break;
		}
	}
}

function buildThings() {
	var list = enumDroid(me, DROID_CONSTRUCT);
	for (var i = 0; i < list.length; ++i) {
		
		var droid = list[i];
		if (checkAndRepair(droid))
			continue;
		
		if (droid.order != DORDER_BUILD) {
			buildThingsWithDroid(droid);
		}
	}
}

function scavBuildDroid(fac, name, body, prop, weapon) {
	if(isDefined(weapon[2]))
		return buildDroid(fac, name, body, prop, null, null, weapon[0], weapon[1], weapon[2]);
	
	else if(isDefined(weapon[1]))
		return buildDroid(fac, name, body, prop, null, null, weapon[0], weapon[1]);
	
	else
		return buildDroid(fac, name, body, prop, null, null, weapon[0]);
}

function produceCrane(fac) {
	if (enumDroid(me, DROID_CONSTRUCT).length > 7)
		return false;
	
	if (random(2)) {
		return buildDroid(fac, "Crane", "B2crane1", "BaBaProp", null, null, "scavCrane1");
	}
	else {
		return buildDroid(fac, "Crane", "B2crane2", "BaBaProp", null, null, "scavCrane2");
	}
}

function produceDroid(fac) {
	if ((countDroid(DROID_CONSTRUCT, me) < 3) || !random(10)) {
		produceCrane(fac)
		return;
	}
	var weapons = [];
	
	if (!random(10)) {
		if (enumDroid(me, DROID_SENSOR).length < MAX_SENSORS) {
			weapons.push("ScavSensor");
			scavBuildDroid(fac, "Sensor", "BusBody", "BaBaProp", weapons);
		}
	}
	else {
		var j = random(templates.length);
		var name = (templates[j][1].indexOf("NEXUS") > -1) ? "Nexus Tank" : "Tank";
		
		for(var x = 1; x < 4; ++x) {
			if(isDefined(templates[j][x]))
				weapons.push(templates[j][x]);
			else
				break;
			
		}
		scavBuildDroid(fac, name, templates[j][0], "BaBaProp", weapons);
	}
}

function produceVtol(fac) {
	var j = random(vtolTemplates.length);
	var weapons = [];
	
	for(var x = 1; x < 4; ++x) {
		if(isDefined(vtolTemplates[j][x]))
			weapons.push(vtolTemplates[j][x]);
		else
			break;
	}
	scavBuildDroid(fac, "Scavenger VTOL", vtolTemplates[j][0], "V-Tol", weapons);
}

function structureReady(struct) {
	return structureIdle(struct) && (struct.status == BUILT);
}

function produceThings() {
	if(atLimits())
		return;
	
	var list = enumStruct(me,factory);
	for (var i = 0; i < list.length; ++i) {
		var fac = list[i];
		
		if (structureReady(fac))
			produceDroid(fac);
	}
	
	list = enumStruct(me,vtolfac);
	for (var i = 0; i < list.length; ++i) {
		var fac = list[i];
		
		if (structureReady(fac))
			produceVtol(fac);
	}
}


/* This function is way too costly on performance
function findNearestTarget(x, y) {
	var minTarget;
	
	for (var i = 0; i < maxPlayers; ++i) {
		if(i === me) //do not target self.
			continue;
		
		var x = findNearest(enumStruct(i), x, y, true);
		if(isDefined(x)) {
			minTarget = x;
			
			//Randomly choose this target
			if(!random(10))
				return minTarget;
		}
	}
	return minTarget;
}
*/

function checkAndRepair(droid) {
	if (isVTOL(droid) || (droid.order == DORDER_BUILD))
		return false;
	
	if (droid.health < 55) {
		if(orderDroid(droid, DORDER_RTR))
			return true;
	}
	return false;
}

function attackWithDroid(droid, target, force) {
	if(checkAndRepair(droid))
		return;
	
	if (droid.droidType == DROID_WEAPON) {
		if ((droid.order != DORDER_ATTACK) || force) {
			var coords = mapLimits(target.x, target.y, 5, 2, 0, 0);
			orderDroidLoc(droid, DORDER_SCOUT, coords[0], coords[1]);
		}
	}
	else if(droid.droidType == DROID_SENSOR) {
		if ((droid.order != DORDER_OBSERVE) || force)
			orderDroidObj(droid, DORDER_OBSERVE, target);
	}
}

//Taken from nullbot v3.06
function vtolArmed(obj, percent) {
	if (obj.type != DROID)
		return;
	
	if (!isVTOL(obj))
		return false;
	
	for (var i = 0; i < obj.weapons.length; ++i)
		if (obj.weapons[i].armed >= percent)
			return true;
		
	return false;
}

//Taken from nullbot v3.06
function vtolReady(droid) {
	if (droid.order == DORDER_ATTACK)
		return false;
	
	if (vtolArmed(droid, 1))
		return true;
	
	if (droid.order != DORDER_REARM) {
		orderDroid(droid, DORDER_REARM);
	}
	
	return false;
}

//Helicopters can only attack things that the scavengers have seen
function helicopterAttack() {
	var list = sortHelicopters();
	for (var i = 0; i < list.length; ++i) {
		var droid = list[i];
		
		var target = rangeStep(droid, true);
		var coords = [];
		
		if(isDefined(target)) {
			coords = mapLimits(target.x, target.y, 5, 2, 0, 0);
		}
		else {
			var xOff = random(2);
			var yOff = random(2);
			xOff = (!xOff) ? -random(10) : random(10);
			yOff = (!yOff) ? -random(10) : random(10);
			coords = mapLimits(droid.x, droid.y, 5, 2, xOff, yOff);
		} 
		if(vtolReady(droid))
			orderDroidLoc(droid, DORDER_SCOUT, coords[0], coords[1]);
	}
}

function sortHelicopters() {
	return enumDroid(me).filter(function(object) {
		return object.propulsion === "V-Tol";
	});
}

function groundAttackStuff() {
	var randomBase = baseInfo[random(baseInfo.length)];
	var attackDroids = enumGroup(randomBase.attackGroup);
	var nexusDroids = enumGroup(randomBase.nexusGroup);
	var target;
	
	if(attackDroids.length > 0)
		target = rangeStep(attackDroids[0], false);
	//var target = findNearestTarget(randomBase.x, randomBase.y);
		
	if(isDefined(target)) {	
		if(attackDroids.length > MIN_ATTACKERS) {
			for (var i = 0; i < attackDroids.length; ++i) {	
				if(isDefined(target)) {
					attackWithDroid(attackDroids[i], target, false);
				}		
			}
		}
		if(nexusDroids.length > MIN_NEXUS) {
			for (var i = 0; i < nexusDroids.length; ++i) {		
				if(isDefined(target)) {
					attackWithDroid(nexusDroids[i], target, true);
				}	
			}
		}
	}
}


function eventAttacked(victim, attacker) {
	if (attacker.player == me) // don't quarrel because of friendly splash damage
		return;
		
	if (victim.type == STRUCTURE) {
		var base = findNearest(baseInfo, victim.x, victim.y, true);
		var list = enumGroup(base.defendGroup);
		
		//Let this base build more defense units then
		if (list.length < MAX_DEFENDERS)
			list = enumGroup(base.attackDroids);
		
		list = list.concat(enumGroup(globalDefendGroup));
		
		for (var i = 0; i < list.length; ++i)
			attackWithDroid(list[i], attacker, true);
		
	}
	else if (victim.type == DROID) {
		if (isVTOL(victim)) {
			orderDroidObj(victim, DORDER_ATTACK, attacker);
			return;
		}
		
		var gr = groupOfTank(victim);
		if (!isDefined(gr)) 
			return;
		
		var list = enumGroup(gr);
		for (var i = 0; i < list.length; ++i)
			if (list[i].id != victim.id)
				attackWithDroid(list[i], attacker, true);
	}
}

function eventDroidBuilt(droid, fac) {
	if (droid.propulsion !== "V-Tol")
		groupAddDroid(needToPickGroup, droid);
	
	queue("reviseGroups", 200);
	queue("produceThings");
}

function eventStructureBuilt(structure, droid) {
	if (structure.stattype == FACTORY) {
		if (!produceCrane(structure))
			produceDroid(structure);
	} 
	else if (structure.stattype == VTOL_FACTORY)
		produceVtol(structure);	
}

function eventGameInit() {
	log("eventGameInit()");
	for (var i = 0; i < templates.length; ++i) {
		
		makeComponentAvailable(templates[i][0], me);
		makeComponentAvailable(templates[i][1], me);
		
		if (isDefined(templates[i][2]))
			makeComponentAvailable(templates[i][2], me);
		
		if (isDefined(templates[i][3]))
			makeComponentAvailable(templates[i][3], me);
		
	}
	for (var i = 0; i < vtolTemplates.length; ++i) {
		
		makeComponentAvailable(vtolTemplates[i][0], me);
		makeComponentAvailable(vtolTemplates[i][1], me);
		
		if (isDefined(vtolTemplates[i][2]))
			makeComponentAvailable(vtolTemplates[i][2], me);
		
		if (isDefined(vtolTemplates[i][3]))
			makeComponentAvailable(vtolTemplates[i][3], me);
		
	}
	makeComponentAvailable("B2crane1", me);
	makeComponentAvailable("scavCrane1", me);
	makeComponentAvailable("B2crane2", me);
	makeComponentAvailable("scavCrane2", me);
	makeComponentAvailable("ScavSensor", me);
	makeComponentAvailable("BaBaProp", me);
	makeComponentAvailable("V-Tol", me);
	
	enableStructure(factory, me);
	enableStructure(vtolfac, me);
	enableStructure(derrick, me);
	enableStructure(gen, me);
	enableStructure(repair, me);
	enableStructure(vtolpad, me);
	
	for (var i = 0; i < defenses.length; ++i)
		enableStructure(defenses[i], me);
}

// respond correctly on unit transfers
function eventObjectTransfer(object, from) {
	logObj(object, "eventObjectTranfer: A droid was transfered to me from player " + from);
	
	//Under no circumstance should this droid be allowed to live
	if((object.health > 100) && !allianceExistsBetween(from, me)) {
		logObj(object, "eventObjectTranfer: Destroying droid to stop CHECK_DROID asserts.");
		removeObject(object, true);
		return;
	}
	
	if (object.type == DROID) {
		eventDroidBuilt(object, null);
	} else {
		eventStructureBuilt(object, null);
	}
}

function eventStartLevel() {
	log("eventStartLevel()");
	
	var list = enumStruct(me, factory);
	for (var i = 0; i < list.length; ++i) {
		var fac = list[i];
		baseInfo[i] = new constructBaseInfo(fac.x, fac.y);
	}
	MAX_SENSORS = baseInfo.length;
	list = enumDroid(me);
	
	for (var i = 0; i < list.length; ++i)
		addDroidToSomeGroup(list[i]);
	
	globalDefendGroup = newGroup();
	needToPickGroup = newGroup();
	
	produceThings();
	setTimer("groundAttackStuff", 2000);
	setTimer("helicopterAttack", 4000);
	setTimer("buildThings", 5000);
	setTimer("produceThings", 10000);
}

